/*
 *
 * Copyright (c) 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Hourglass.
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation is hereby granted without fee, provided that the
 * above copyright notice and this permission/disclaimer notice is
 * retained in all copies or modified versions, and that both notices
 * appear in supporting documentation.  THE COPYRIGHT HOLDERS PROVIDE
 * THIS SOFTWARE "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE COPYRIGHT
 * HOLDERS DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER
 * RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu
 * any improvements that they make and grant redistribution rights to the
 * University of Utah.
 * 
 * Author: John Regehr (regehr@cs.utah.edu)
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
//#include <math.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <assert.h>

#include "config.h"

// #define GAP_HIST 1

typedef signed long long int cycle_time;
typedef double us_time;
typedef double sec_time;

extern int need_rtc_timers;
extern unsigned int duration;
extern int tracing;
extern int num_threads;
extern double MHZ;
extern struct thrd_data_t *thrd_data;
extern struct trace_rec_t *trace_recs;
extern long current_record;
extern cycle_time start_time, end_time;
extern cycle_time GAP;

extern cycle_time trace_start;
extern cycle_time trace_end;
extern int print_cycles;

// FIXME - this is hardware dependent
// it means that on average it takes 200ns to
// run the poll loop, and this has to be 
#define POLL_CORRECTION ((us_time)(0.2/2))

#ifdef GAP_HIST
extern int *gap_hist;
#define NUM_GAP_HIST_BUCKETS 1000
#define GAP_HIST_DIVISOR 10
#endif

#define TRUE 1
#define FALSE 0

/*
 * number of microseconds to wait before threads start doing things;
 * this helps ensure that everyone sets priorty, creates reservations,
 * etc. before the high-priority threads soak up all CPU time
 */
#define INITIAL_DELAY 5000

enum thrd_work_t { 
  CPU = 121, 
  PERIODIC,
  CPU_PERIODIC,
  CPU_SCAN,
  CPU_SCAN_YIELD,
  CPU_YIELD,
  LAT
};

enum thrd_pri_t { 
  IDLE = 222, 
  LOW, 
  NORMAL, 
  HIGH, 
  HIGHEST, 
  RTLOW, 
  RTMED, 
  RTHIGH, 
  TIMER_PRI, 
  MAIN_PRI 
};

enum thrd_status_t { 
  INIT = 333, 
  READY, 
  GO, 
  STOP, 
  DONE 
};

#define MAXSTRLEN 512
#define LAT_SLOTS 50000
#define NUM_TRACE_RECORDS 300000
#define MAX_SPIN 500

/* argh - usleep of <1-2ms usually busy-waits */
#define SMALL_SLEEP 5000

#define ASSERT(x) assert(x)

struct timer_funcs {
  void (*timed_sleep) (int, us_time);
  void (*init_timers) (void);
  void (*deinit_timers) (void);
};

#ifdef WITH_RTCTIMERS
extern const struct timer_funcs rtc_timer_funcs;
#endif

#ifdef WITH_HRTIMERS
extern const struct timer_funcs hr_timer_funcs;
#endif

extern const struct timer_funcs unix_timer_funcs;

struct thrd_data_t {
  // generic stuff
  volatile enum thrd_status_t status;
  enum thrd_work_t work; 
  enum thrd_pri_t pri;
  pthread_t thrd;
  pthread_t monitor_thrd;
  int id;
  int log;
  const struct timer_funcs *timers;
  long long int workdone;

  us_time s_time, f_time;

  int *scanmem;
  int memsize;

  // periodic thread stuff
  us_time app_period, app_amount;
  int hits, misses;

  // latency test stuff
  cycle_time *lat_slots;
  int lat_slot;

  // reservation stuff
  us_time res_period, res_amount;
  int res_soft;

  // execution trace stuff
  cycle_time last_exec;
  cycle_time exec_start;
  cycle_time thread_time;
  int reserved_record;
  const char *bossa_policy;
  int *bossa_args;
};

struct trace_rec_t {
  int id;
  cycle_time start, end;
};

/*
 * rdtsc: this works with gcc on Pentium and Athlon class machines
 *
 * a bit of documentation since this function seems to often be
 * misunderstood:
 *
 * rdtsc is an instruction and "=A" (d) tells gcc that we're
 * outputting a 64-bit quantity, stored in eax:edx, and putting
 * it into variable d
 *
 * when optimization is turned on, gcc does the right thing: it
 * inlines this function and (properly) eliminates storage location d 
 */
static inline cycle_time rdtsc (void)
{
  cycle_time d;
  __asm__ __volatile__ ("rdtsc" : "=A" (d) );
  return d;
}

static inline sec_time cycle_to_sec (cycle_time c)
{
  return c / (sec_time)(MHZ * 1000000);
}

static inline cycle_time sec_to_cycle (sec_time s)
{
  return (cycle_time)s * (cycle_time)(MHZ * 1000000);
}

static inline cycle_time us_to_cycle (us_time us)
{
  return (cycle_time)us * (cycle_time)MHZ;
}

static inline us_time cycle_to_us (cycle_time c)
{
  return c / (us_time)MHZ;
}

static inline us_time sec_to_us (sec_time s)
{
  return (us_time)s * 1000000;
}

static inline sec_time us_to_sec (us_time us)
{
  return (sec_time)(us/1000000);
}

extern void *xmalloc (size_t size);

extern void set_priority (enum thrd_pri_t pri);
extern void lock_pages (void);
extern double find_mhz (void);

extern void *thrd_func (void *arg);

extern void *monitor_thrd_func (void *arg);

extern void print_all_results (void);

extern void parse_args (int argc, char *argv[]);

/* these routines snagged from Wine */

static inline int my_InterlockedExchangeAdd (long *dest, long incr)
{
    long ret;
    __asm__ __volatile__( "lock; xaddl %0,(%1)"
                          : "=r" (ret) : "r" (dest), "0" (incr) : "memory" );
    return ret;
}

static inline int my_InterlockedIncrement (long *dest)
{
    return my_InterlockedExchangeAdd (dest, 1) + 1;
}

extern void make_res (us_time amount, 
		      us_time period,
		      int which);

extern void delete_res (int which);

extern pid_t hourglass_pid;

#define mymin(x,y) ((x)<(y)?(x):(y))
